__title__ = 'dbus_next'
__description__ = 'A zero-dependency DBus library for Python with asyncio support'
__url__ = 'https://github.com/altdesktop/python-dbus-next'
__version__ = '0.2.2'
__author__ = 'Tony Crisci'
__author_email__ = 'tony@dubstepdish.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2019 Tony Crisci'
